@@ -23,6 +23,10 @@
 
 #define MAX_RETRIES 25
 
+#ifdef GAMIN_DEBUG_API
+int debug_reqno = -1;
+void *debug_userData = NULL;
+#endif
 int FAMErrno = 0;
 
 enum {
@@ -46,6 +50,11 @@ const char *FamErrlist[] = {
     NULL
 };
 
+#ifdef GAMIN_DEBUG_API
+int FAMDebug(FAMConnection *fc, const char *filename, FAMRequest * fr,
+             void *userData);
+#endif
+
 #ifdef TEST_DEBUG
 static char *
 gamin_dump_event(FAMEvent *event) {
@@ -71,12 +80,20 @@ gamin_dump_event(FAMEvent *event) {
 }
 #endif
 
+void gam_show_debug(void);
+
+void
+gam_show_debug(void) {
+}
+
 /************************************************************************
  *									*
  *			Path for the socket connection			*
  *									*
  ************************************************************************/
 
+static char user_name[100] = "";
+
 /**
  * gamin_get_user_name:
  *
@@ -84,17 +101,22 @@ gamin_dump_event(FAMEvent *event) {
  *
  * Returns a new string or NULL in case of error.
  */
-static char *
+static const char *
 gamin_get_user_name(void)
 {
-	struct passwd *pw;
+    struct passwd *pw;
 
-	pw = getpwuid(getuid());
+    if (user_name[0] != 0)
+        return (user_name);
 
-	if (pw != NULL)
-		return (strdup(pw->pw_name));
+    pw = getpwuid(getuid());
 
-	return (NULL);
+    if (pw != NULL) {
+	strncpy(user_name, pw->pw_name, 99);
+	user_name[99] = 0;
+    }
+
+    return(user_name);
 }
 
 /**
@@ -110,7 +132,7 @@ static char *
 gamin_get_socket_path(void)
 {
     const char *fam_client_id;
-    char *user;
+    const char *user;
     char *ret;
     char path[MAXPATHLEN + 1];
 
@@ -132,7 +154,6 @@ gamin_get_socket_path(void)
 #endif
     path[MAXPATHLEN] = 0;
     ret = strdup(path);
-    free(user);
     return (ret);
 }
 
@@ -223,7 +244,7 @@ gamin_check_secure_dir(void)
     /*
      * all checks on existing dir seems okay
      */
-    gam_debug(DEBUG_INFO, "Reusing socket directory %s\n", dir);
+    GAM_DEBUG(DEBUG_INFO, "Reusing socket directory %s\n", dir);
     free(dir);
     return(1);
 
@@ -242,7 +263,7 @@ unsafe:
 	    return(-1);
 	}
     }
-    gam_debug(DEBUG_INFO, "Removed %s\n", dir);
+    GAM_DEBUG(DEBUG_INFO, "Removed %s\n", dir);
     free(dir);
     return(0);
 }
@@ -434,10 +455,13 @@ gamin_data_available(int fd)
     struct timeval tv;
     int avail;
 
-    if (fd < 0)
+    if (fd < 0) {
+        GAM_DEBUG(DEBUG_INFO, "gamin_data_available wrong fd %d\n", fd);
         return (-1);
+    }
+    GAM_DEBUG(DEBUG_INFO, "Checking data available on %d\n", fd);
 
-  retry:
+retry:
     FD_ZERO(&read_set);
     FD_SET(fd, &read_set);
     tv.tv_sec = 0;
@@ -507,6 +531,21 @@ gamin_send_request(GAMReqType type, int fd, const char *filename,
     GAMPacket req;
     int ret;
 
+#ifdef GAMIN_DEBUG_API
+    if (type == GAM_REQ_DEBUG) {
+        len = strlen(filename);
+        if (len > MAXPATHLEN) {
+	    FAMErrno = FAM_FILE;
+            return (-1);
+	}
+        reqnum = gamin_data_get_reqnum(data, filename, (int) type, userData);
+        if (reqnum < 0) {
+	    FAMErrno = FAM_ARG;
+            return (-1);
+	}
+        reqnum = fr->reqnum;
+    } else
+#endif
     if (filename == NULL) {
         len = 0;
         reqnum = fr->reqnum;
@@ -541,6 +580,10 @@ gamin_send_request(GAMReqType type, int fd, const char *filename,
     req.version = GAM_PROTO_VERSION;
     req.seq = reqnum;
     req.type = (unsigned short) type;
+    if ((type == GAM_REQ_DIR) && (gamin_data_get_exists(data) == 0)) {
+        req.type |= GAM_OPT_NOEXISTS;
+    }
+        
     req.pathlen = len;
     if (len > 0)
         memcpy(&req.path[0], filename, len);
@@ -609,23 +652,23 @@ retry:
         if (errno == EINTR)
             goto retry;
 
-        gam_debug(DEBUG_INFO, "Failed to read credentials byte on %d\n", fd);
+        GAM_DEBUG(DEBUG_INFO, "Failed to read credentials byte on %d\n", fd);
         goto failed;
     }
 
     if (buf != '\0') {
-        gam_debug(DEBUG_INFO, "Credentials byte was not nul on %d\n", fd);
+        GAM_DEBUG(DEBUG_INFO, "Credentials byte was not nul on %d\n", fd);
         goto failed;
     }
 #ifdef HAVE_CMSGCRED
     if (cmsg->cmsg_len < sizeof(cmsgmem) || cmsg->cmsg_type != SCM_CREDS) {
-        gam_debug(DEBUG_INFO,
+        GAM_DEBUG(DEBUG_INFO,
                   "Message from recvmsg() was not SCM_CREDS\n");
         goto failed;
     }
 #endif
 
-    gam_debug(DEBUG_INFO, "read credentials byte\n");
+    GAM_DEBUG(DEBUG_INFO, "read credentials byte\n");
 
     {
 #ifdef SO_PEERCRED
@@ -638,7 +681,7 @@ retry:
             c_uid = cr.uid;
             c_gid = cr.gid;
         } else {
-            gam_debug(DEBUG_INFO,
+            GAM_DEBUG(DEBUG_INFO,
                       "Failed to getsockopt() credentials on %d, returned len %d/%d\n",
                       fd, cr_len, (int) sizeof(cr));
             goto failed;
@@ -652,19 +695,19 @@ retry:
         c_uid = cred->cmcred_euid;
         c_gid = cred->cmcred_groups[0];
 #else /* !SO_PEERCRED && !HAVE_CMSGCRED */
-        gam_debug(DEBUG_INFO,
+        GAM_DEBUG(DEBUG_INFO,
                   "Socket credentials not supported on this OS\n");
         goto failed;
 #endif
     }
 
     if (s_uid != c_uid) {
-        gam_debug(DEBUG_INFO,
+        GAM_DEBUG(DEBUG_INFO,
                   "Credentials check failed: s_uid %d, c_uid %d\n",
                   (int) s_uid, (int) c_uid);
         goto failed;
     }
-    gam_debug(DEBUG_INFO,
+    GAM_DEBUG(DEBUG_INFO,
               "Credentials: s_uid %d, c_uid %d, c_gid %d, c_pid %d\n",
               (int) s_uid, (int) c_uid, (int) c_gid, (int) c_pid);
     gamin_data_done_auth(conn);
@@ -680,6 +723,7 @@ failed:
  * gamin_read_data:
  * @conn: the connection
  * @fd: the file descriptor for the socket
+ * @block: allow blocking
  *
  * Read the available data on the file descriptor. This is a potentially
  * blocking operation.
@@ -687,7 +731,7 @@ failed:
  * Return 0 in case of success, -1 in case of error.
  */
 static int
-gamin_read_data(GAMDataPtr conn, int fd)
+gamin_read_data(GAMDataPtr conn, int fd, int block)
 {
     int ret;
     char *data;
@@ -695,42 +739,49 @@ gamin_read_data(GAMDataPtr conn, int fd)
 
     ret = gamin_data_need_auth(conn);
     if (ret == 1) {
-        gam_debug(DEBUG_INFO, "Client need auth %d\n", fd);
+        GAM_DEBUG(DEBUG_INFO, "Client need auth %d\n", fd);
         if (gamin_check_cred(conn, fd) < 0) {
 	    return (-1);
 	}
-	ret = gamin_data_available(fd);
-	if (ret < 0)
-	    return(-1);
-	if (ret == 0)
-	    return(0);
+	if (!block) {
+	    ret = gamin_data_available(fd);
+	    if (ret < 0)
+	        return(-1);
+	    if (ret == 0)
+	        return(0);
+	}
     } else if (ret != 0) {
-	return (-1);
+        goto error;
     }
     ret = gamin_data_get_data(conn, &data, &size);
     if (ret < 0) {
-        return (-1);
+        GAM_DEBUG(DEBUG_INFO, "Failed getting connection data\n");
+        goto error;
     }
 retry:
     ret = read(fd, (char *) data, size);
     if (ret < 0) {
-        if (errno == EINTR)
+        if (errno == EINTR) {
+	    GAM_DEBUG(DEBUG_INFO, "client read() call interrupted\n");
             goto retry;
+	}
         gam_error(DEBUG_INFO, "failed to read() from server connection\n");
-        return (-1);
+        goto error;
     }
     if (ret == 0) {
         gam_error(DEBUG_INFO, "end from FAM server connection\n");
-        return (-1);
+        goto error;
     }
-    gam_debug(DEBUG_INFO, "read %d bytes from server\n", ret);
+    GAM_DEBUG(DEBUG_INFO, "read %d bytes from server\n", ret);
 
     if (gamin_data_conn_data(conn, ret) < 0) {
         gam_error(DEBUG_INFO, "Failed to process %d bytes from server\n",
                   ret);
-        return (-1);
+        goto error;
     }
     return (0);
+error:
+    return(-1);
 }
 
 /**
@@ -761,7 +812,10 @@ gamin_resend_request(int fd, GAMReqType type, const char *filename,
     req.len = (unsigned short) tlen;
     req.version = GAM_PROTO_VERSION;
     req.seq = reqnum;
-    req.type = (unsigned short) type;
+    /* GAM_OPT_NOEXISTS to avoid filling up the connection with
+       events we don't need and discard */
+    req.type = (unsigned short) (type | GAM_OPT_NOEXISTS);
+    /* req.type = (unsigned short) type; */
     req.pathlen = len;
     if (len > 0)
         memcpy(&req.path[0], filename, len);
@@ -1141,7 +1195,7 @@ FAMMonitorCollection(FAMConnection * fc, const char *filename,
  *
  * Read the next event, possibly blocking on input.
  *
- * Returns 0 in case of success and -1 in case of error.
+ * Returns 1 in case of success and -1 in case of error.
  */
 int
 FAMNextEvent(FAMConnection * fc, FAMEvent * fe)
@@ -1170,7 +1224,7 @@ FAMNextEvent(FAMConnection * fc, FAMEvent * fe)
     }
 
     if (!gamin_data_event_ready(conn)) {
-        if (gamin_read_data(conn, fc->fd) < 0) {
+        if (gamin_read_data(conn, fc->fd, 1) < 0) {
 	    gamin_try_reconnect(conn, fc->fd);
 	    FAMErrno = FAM_CONNECT;
 	    return (-1);
@@ -1216,6 +1270,9 @@ FAMPending(FAMConnection * fc)
 
     GAM_DEBUG(DEBUG_INFO, "FAMPending(fd = %d)\n", fc->fd);
 
+    if (gamin_data_event_ready(conn))
+      return (1);
+
     /*
      * make sure we won't block if reading
      */
@@ -1223,7 +1280,7 @@ FAMPending(FAMConnection * fc)
     if (ret < 0)
         return (-1);
     if (ret > 0) {
-        if (gamin_read_data(conn, fc->fd) < 0) {
+        if (gamin_read_data(conn, fc->fd, 0) < 0) {
 	    gamin_try_reconnect(conn, fc->fd);
 	}
     }
@@ -1241,7 +1298,7 @@ FAMPending(FAMConnection * fc)
  * Returns 0 in case of success and -1 in case of error.
  */
 int
-FAMCancelMonitor(FAMConnection * fc, FAMRequest * fr)
+FAMCancelMonitor(FAMConnection * fc, const FAMRequest * fr)
 {
     GAMDataPtr conn;
     int ret;
@@ -1273,7 +1330,7 @@ FAMCancelMonitor(FAMConnection * fc, FAMRequest * fr)
      * send destruction message to the server
      */
     ret = gamin_send_request(GAM_REQ_CANCEL, fc->fd, NULL,
-                             fr, NULL, fc->client, 0);
+                             (FAMRequest *) fr, NULL, fc->client, 0);
 
     if (ret != 0) {
         FAMErrno = FAM_CONNECT;
@@ -1313,3 +1370,82 @@ int FAMResumeMonitor(FAMConnection *fc, const FAMRequest *fr) {
     FAMErrno = FAM_UNIMPLEM;
     return (-1);
 }
+
+/**
+ * FAMNoExists:
+ * @fc: pointer to a connection structure.
+ *
+ * Specific extension for the core FAM API where Exists event are not
+ * propagated on directory monitory listing startup. This speeds up
+ * watching large directories but can introduce a mismatch between the FAM
+ * view of the directory and the program own view.
+ *
+ * Returns 0 in case of success and -1 in case of error.
+ */
+int FAMNoExists(FAMConnection *fc) {
+    int ret;
+    GAMDataPtr conn;
+
+    if (fc == NULL) {
+	GAM_DEBUG(DEBUG_INFO, "FAMNoExists() arg error\n");
+        FAMErrno = FAM_ARG;
+        return (-1);
+    }
+    conn = fc->client;
+
+    ret = gamin_data_no_exists(conn);
+    if (ret < 0) {
+	GAM_DEBUG(DEBUG_INFO, "FAMNoExists() arg error\n");
+        FAMErrno = FAM_ARG;
+        return(-1);
+    }
+    return(0);
+}
+
+#ifdef GAMIN_DEBUG_API
+/**
+ * FAMDebug:
+ * @fc: pointer to a connection structure.
+ * @filename: file name.
+ * @fr: pointer to the request structure.
+ * @userdata: user provided data for the callbacks
+ *
+ * Specific extension for the core FAM API to request debug informations
+ *
+ * Returns 0 in case of success and -1 in case of error.
+ */
+int
+FAMDebug(FAMConnection *fc, const char *filename, FAMRequest * fr,
+         void *userData)
+{
+    int ret;
+
+    if ((fc == NULL) || (filename == NULL) || (fr == NULL)) {
+	GAM_DEBUG(DEBUG_INFO, "FAMDebug() arg error\n");
+        FAMErrno = FAM_ARG;
+        return (-1);
+    }
+    if (strlen(filename) >= MAXPATHLEN) {
+        FAMErrno = FAM_FILE;
+        return (-1);
+    }
+    if ((fc->fd < 0) || (fc->client == NULL)) {
+	GAM_DEBUG(DEBUG_INFO, "FAMDebug() arg error\n");
+        FAMErrno = FAM_ARG;
+        return (-1);
+    }
+
+    GAM_DEBUG(DEBUG_INFO, "FAMDebug(%s)\n", filename);
+
+    /*
+     * send debug message to the server
+     */
+    ret = gamin_send_request(GAM_REQ_DEBUG, fc->fd, filename,
+			     fr, userData, fc->client, 0);
+    if (debug_reqno == -1) {
+        debug_reqno = fr->reqnum;
+	debug_userData = userData;
+    }
+    return(ret);
+}
+#endif
