@@ -12,9 +12,15 @@
 #include "gam_data.h"
 #include "gam_protocol.h"
 #include "gam_error.h"
+#include "config.h"
 
 #define FAM_EVENT_SIZE (sizeof(FAMEvent))
 
+#ifdef GAMIN_DEBUG_API
+extern int debug_reqno;
+extern void *debug_userData;
+#endif
+
 typedef enum {
     REQ_NONE = 0,               /* not set */
     REQ_INIT = 1,               /* set but no event yet */
@@ -32,7 +38,8 @@ typedef enum {
 struct GAMData {
     int reqno;                  /* counter for the requests */
     int auth;			/* did authentication took place */
-    int restarted;			/* did authentication took place */
+    int restarted;		/* did authentication took place */
+    int noexist;		/* no EXISTS activated */
 
     int evn_ready;              /* do we have a full event ready */
     int evn_read;               /* how many bytes were read for the event */
@@ -503,6 +510,17 @@ gamin_data_conn_event(GAMDataPtr conn, GAMPacketPtr evn)
     if ((conn == NULL) || (evn == NULL))
         return (-1);
 
+#ifdef GAMIN_DEBUG_API
+    if (evn->type >= 50) {
+        GAM_DEBUG(DEBUG_INFO, "Got Debug Event: type %d, seq %d\n",
+	          evn->type, evn->seq);
+	conn->evn_ready = 1;
+	conn->evn_reqnum = debug_reqno;
+	conn->evn_userdata = debug_userData;
+        return(1);
+    }
+#endif
+    
     /* Check the event number */
     req = gamin_data_get_req(conn, evn->seq);
     if (req == NULL) {
@@ -757,3 +775,39 @@ gamin_data_get_request(GAMDataPtr conn, const char *filename, int type,
         return (-1);
     return (req->reqno);
 }
+
+/**
+ * gamin_data_no_exists:
+ * @conn:  a connection data structure
+ *
+ * Switch the connection to a mode where no exists are sent on directory
+ * monitoting startup.
+ *
+ * Returns 0 in case of success and -1 in case of error.
+ */
+int
+gamin_data_no_exists(GAMDataPtr conn)
+{
+    if (conn == NULL)
+        return (-1);
+    conn->noexist = 1;
+    return(0);
+}
+
+/**
+ * gamin_data_get_exists:
+ * @conn:  a connection data structure
+ *
+ * Get the EXISTS flag for the connection
+ *
+ * Returns 0 or 1 in case or -1 in case of error.
+ */
+int
+gamin_data_get_exists(GAMDataPtr conn)
+{
+    if (conn == NULL)
+        return (-1);
+    if (conn->noexist)
+        return(0);
+    return(1);
+}
