@@ -25,6 +25,7 @@
 #include <unistd.h>
 #include <time.h>
 #include <errno.h>
+#include <string.h>
 #include <glib.h>
 #include "fam.h"
 #include "gam_error.h"
@@ -32,9 +33,12 @@
 #include "gam_poll.h"
 #include "gam_event.h"
 #include "gam_server.h"
+#include "gam_protocol.h"
 #include "gam_event.h"
 #include "gam_excludes.h"
 
+/* #define VERBOSE_POLL */
+
 #define DEFAULT_POLL_TIMEOUT 3
 
 #define FLAG_NEW_NODE 1 << 5
@@ -46,6 +50,7 @@
 #define MON_MISSING	1 << 0	/* The resource is missing */
 #define MON_NOKERNEL	1 << 1  /* file(system) not monitored by the kernel */
 #define MON_BUSY	1 << 2  /* Too busy to be monitored by the kernel */
+#define MON_WRONG_TYPE	1 << 3  /* Expecting a directory and got a file */
 
 typedef struct {
     struct stat sbuf;		/* The stat() informations in last check */
@@ -56,25 +61,22 @@ typedef struct {
 } GamPollData;
 
 static GamTree *tree = NULL;
-
 static GList *new_subs = NULL;
-
-G_LOCK_DEFINE_STATIC(new_subs);
-
-static GList *removed_subs = NULL;
-
-G_LOCK_DEFINE_STATIC(removed_subs);
-
 static GList *missing_resources = NULL;
-
-G_LOCK_DEFINE_STATIC(missing_resources);
-
+static GList *all_resources = NULL;
 static GamPollHandler dir_handler = NULL;
 static GamPollHandler file_handler = NULL;
 
 static int poll_mode = 0;
 
-static time_t current_time;	/* a cache for time() informations */
+static time_t current_time = 0;	/* a cache for time() informations */
+
+static GaminEventType poll_file(GamNode * node);
+
+static int
+gam_errno(void) {
+    return(errno);
+}
 
 static void
 trigger_dir_handler(const char *path, gboolean added)
@@ -90,52 +92,75 @@ trigger_file_handler(const char *path, gboolean added)
         (*file_handler) (path, added);
 }
 
-static void
+static int
 node_add_subscription(GamNode * node, GamSubscription * sub)
 {
-    const char *path;
-
     if ((node == NULL) || (sub == NULL))
-        return;
+        return(-1);
 
-    gam_debug(DEBUG_INFO, "node_add_subscription(%s)\n", node->path);
+    if ((node->path == NULL) || (node->path[0] != '/'))
+        return(-1);
+
+    GAM_DEBUG(DEBUG_INFO, "node_add_subscription(%s)\n", node->path);
     gam_node_add_subscription(node, sub);
 
-    path = gam_node_get_path(node);
-    if (gam_exclude_check(path)) {
-	gam_debug(DEBUG_INFO, "  gam_exclude_check: true\n");
-        return;
+    if (gam_exclude_check(node->path)) {
+	GAM_DEBUG(DEBUG_INFO, "  gam_exclude_check: true\n");
+	if (gam_node_get_data(node) == NULL)
+	    poll_file(node);
+	    
+	gam_poll_add_missing(node);
+        return(0);
     }
 
     if (gam_node_is_dir(node))
-        trigger_dir_handler(gam_node_get_path(node), TRUE);
+        trigger_dir_handler(node->path, TRUE);
     else
-        trigger_file_handler(gam_node_get_path(node), TRUE);
+        trigger_file_handler(node->path, TRUE);
 
+    return(0);
 }
 
-static void
+static int
 node_remove_subscription(GamNode * node, GamSubscription * sub)
 {
     const char *path;
 
     if ((node == NULL) || (sub == NULL))
-        return;
+        return(-1);
 
-    gam_debug(DEBUG_INFO, "node_remove_subscription(%s)\n", node->path);
+    if ((node->path == NULL) || (node->path[0] != '/'))
+        return(-1);
+
+    GAM_DEBUG(DEBUG_INFO, "node_remove_subscription(%s)\n", node->path);
 
     gam_node_remove_subscription(node, sub);
 
     path = gam_node_get_path(node);
     if (gam_exclude_check(path)) {
-	gam_debug(DEBUG_INFO, "  gam_exclude_check: true\n");
-        return;
+	GAM_DEBUG(DEBUG_INFO, "  gam_exclude_check: true\n");
+        return(0);
     }
 
-    if (gam_node_is_dir(node))
-        trigger_dir_handler(gam_node_get_path(node), FALSE);
-    else
-        trigger_file_handler(gam_node_get_path(node), FALSE);
+    /* DNotify makes our life miserable here */
+    if (gam_subscription_is_dir(sub)) {
+	if (gam_node_is_dir(node))
+	    trigger_dir_handler(path, FALSE);
+	else {
+	    char *dir;
+
+	    dir = g_path_get_dirname(path);
+	    trigger_file_handler(dir, FALSE);
+	    g_free(dir);
+	}
+    } else {
+	if (gam_node_is_dir(node))
+	    trigger_dir_handler(path, FALSE);
+	else
+	    trigger_file_handler(path, FALSE);
+    }
+
+    return(0);
 }
 
 static GamPollData *
@@ -143,11 +168,22 @@ gam_poll_data_new(const char *path)
 {
     GamPollData *data;
 
-    data = g_new0(GamPollData, 1);
+    if ((path == NULL) || (path[0] != '/'))
+        return(NULL);
+    data = g_new(GamPollData, 1);
+    if (data == NULL)
+        return(NULL);
+    memset(data, 0, sizeof(GamPollData));
     data->path = g_strdup(path);
 
     data->flags = 0;
+    if (current_time == 0)
+        current_time = time(NULL);
     data->lasttime = current_time;
+#ifdef ST_MTIM_NSEC
+    data->sbuf.st_mtim.tv_sec = current_time;
+#endif
+    data->sbuf.st_mtime = current_time;
     data->checks = 0;
 
     return data;
@@ -161,15 +197,10 @@ gam_poll_emit_event(GamNode * node, GaminEventType event,
     GamNode *parent;
     GamPollData *data;
     GList *subs;
+    int is_dir_node = gam_node_is_dir(node);
 
-    /* we only emit CREATED, DELETED and EXISTS for directories */
-    if (gam_node_is_dir(node) &&
-        !(event == GAMIN_EVENT_CREATED ||
-          event == GAMIN_EVENT_DELETED || event == GAMIN_EVENT_EXISTS))
-        return;
-
-    gam_debug(DEBUG_INFO, "Poll: emit events for %s\n",
-              gam_node_get_path(node));
+    GAM_DEBUG(DEBUG_INFO, "Poll: emit events %d for %s\n",
+              event, gam_node_get_path(node));
     subs = gam_node_get_subscriptions(node);
     if (subs)
         subs = g_list_copy(subs);
@@ -191,6 +222,7 @@ gam_poll_emit_event(GamNode * node, GaminEventType event,
         for (l = subs; l; l = l->next) {
             GamSubscription *sub = l->data;
             GaminEventType new_event = event;
+	    GList *tmp;
 
             if (g_list_find(exist_subs, sub)) {
                 if ((data) && (!(data->flags & MON_MISSING)))
@@ -199,11 +231,16 @@ gam_poll_emit_event(GamNode * node, GaminEventType event,
                     continue;
             }
 
-            gam_server_emit_event(gam_node_get_path(node),
-                                  new_event, g_list_append(NULL, sub));
+            tmp = g_list_prepend(NULL, sub);
+            gam_server_emit_event(gam_node_get_path(node), is_dir_node,
+                                  new_event, tmp, 0);
+	    g_list_free(tmp);
+
         }
     } else {
-        gam_server_emit_event(gam_node_get_path(node), event, subs);
+        
+        gam_server_emit_event(gam_node_get_path(node), is_dir_node, event,
+	                      subs, 0);
     }
 
     g_list_free(subs);
@@ -216,6 +253,55 @@ gam_poll_data_destroy(GamPollData * data)
     g_free(data);
 }
 
+/**
+ * gam_poll_delist_node:
+ * @node: the node to delist
+ *
+ * This function is called when kernel monitoring for a node should
+ * be turned off.
+ */
+static void
+gam_poll_delist_node(GamNode * node) {
+    GList *subs;
+
+    GAM_DEBUG(DEBUG_INFO, "gam_poll_delist_node %s\n",
+              gam_node_get_path(node));
+    subs = gam_node_get_subscriptions(node);
+
+    while (subs != NULL) {
+	if (gam_node_is_dir(node))
+	    trigger_dir_handler(gam_node_get_path(node), FALSE);
+	else
+	    trigger_file_handler(gam_node_get_path(node), FALSE);
+	subs = subs->next;
+    }
+}
+
+/**
+ * gam_poll_relist_node:
+ * @node: the node to delist
+ *
+ * This function is called when kernel monitoring for a node should
+ * be turned on (again).
+ */
+static void
+gam_poll_relist_node(GamNode * node) {
+    GList *subs;
+
+    GAM_DEBUG(DEBUG_INFO, "gam_poll_relist_node %s\n",
+              gam_node_get_path(node));
+    subs = gam_node_get_subscriptions(node);
+
+    while (subs != NULL) {
+	if (gam_node_is_dir(node))
+	    trigger_dir_handler(gam_node_get_path(node), TRUE);
+	else
+	    trigger_file_handler(gam_node_get_path(node), TRUE);
+	subs = subs->next;
+    }
+}
+
+
 static GaminEventType
 poll_file(GamNode * node)
 {
@@ -226,13 +312,13 @@ poll_file(GamNode * node)
     const char *path;
 
     path = gam_node_get_path(node);
-    gam_debug(DEBUG_INFO, "Poll: poll_file for %s called\n", path);
+    GAM_DEBUG(DEBUG_INFO, "Poll: poll_file for %s called\n", path);
 
     data = gam_node_get_data(node);
     if (data == NULL) {
         char real[PATH_MAX];
 
-        gam_debug(DEBUG_INFO, "Poll: poll_file : new\n");
+        GAM_DEBUG(DEBUG_INFO, "Poll: poll_file : new\n");
         realpath(path, real);
         data = gam_poll_data_new(real);
 
@@ -242,28 +328,33 @@ poll_file(GamNode * node)
         stat_ret = stat(data->path, &sbuf);
 	if (stat_ret != 0)
 	    data->flags |= MON_MISSING;
+	else
+	    gam_node_set_is_dir(node, (S_ISDIR(sbuf.st_mode) != 0));
 	if (gam_exclude_check(path))
 	    data->flags |= MON_NOKERNEL;
-        data->sbuf = sbuf;
+        memcpy(&(data->sbuf), &(sbuf), sizeof(sbuf));
 
         if (stat_ret == 0)
             return 0;
         else
             return GAMIN_EVENT_DELETED;
     }
-    gam_debug(DEBUG_INFO, " at %d delta %d : %d\n", current_time,
+    GAM_DEBUG(DEBUG_INFO, " at %d delta %d : %d\n", current_time,
               current_time - data->lasttime, data->checks);
 
     event = 0;
 
-    if (stat(data->path, &sbuf) != 0) {
-        if ((errno == ENOENT) && (!(data->flags & MON_MISSING))) {
+    stat_ret = stat(data->path, &sbuf);
+    if (stat_ret != 0) {
+        if ((gam_errno() == ENOENT) && (!(data->flags & MON_MISSING))) {
             /* deleted */
             data->flags = MON_MISSING;
-            event = GAMIN_EVENT_DELETED;
+
 	    if (gam_node_get_subscriptions(node) != NULL) {
+		gam_poll_delist_node(node);
 	        gam_poll_add_missing(node);
 	    }
+            event = GAMIN_EVENT_DELETED;
         }
     } else if (data->flags & MON_MISSING) {
         /* created */
@@ -277,8 +368,8 @@ poll_file(GamNode * node)
                (data->sbuf.st_ctim.tv_nsec != sbuf.st_ctim.tv_nsec)) {
         event = GAMIN_EVENT_CHANGED;
     } else {
-	gam_debug(DEBUG_INFO, "Poll: poll_file %s unchanged\n", path);
-	gam_debug(DEBUG_INFO, "%d %d : %d %d\n", data->sbuf.st_mtim.tv_sec,
+	GAM_DEBUG(DEBUG_INFO, "Poll: poll_file %s unchanged\n", path);
+	GAM_DEBUG(DEBUG_INFO, "%d %d : %d %d\n", data->sbuf.st_mtim.tv_sec,
 	          data->sbuf.st_mtim.tv_nsec, sbuf.st_mtim.tv_sec,
 		  sbuf.st_mtim.tv_nsec);
 #else
@@ -286,11 +377,17 @@ poll_file(GamNode * node)
                (data->sbuf.st_size != sbuf.st_size) ||
                (data->sbuf.st_ctime != sbuf.st_ctime)) {
         event = GAMIN_EVENT_CHANGED;
-	gam_debug(DEBUG_INFO, "%d : %d\n", data->sbuf.st_mtime,
+	GAM_DEBUG(DEBUG_INFO, "%d : %d\n", data->sbuf.st_mtime,
 		  sbuf.st_mtime);
 #endif
     }
 
+    /*
+     * TODO: handle the case where a file/dir is removed and replaced by 
+     *       a dir/file
+     */
+    if (stat_ret == 0)
+	gam_node_set_is_dir(node, (S_ISDIR(sbuf.st_mode) != 0));
     data->sbuf = sbuf;
 
     /*
@@ -320,92 +417,67 @@ poll_file(GamNode * node)
 
     if ((data->checks >= 4) && (!(data->flags & MON_BUSY))) {
 	if (gam_node_get_subscriptions(node) != NULL) {
-#if 0
-	    fprintf(stderr, "switching %s back to polling\n", path);
-#endif
+	    GAM_DEBUG(DEBUG_INFO, "switching %s back to polling\n", path);
 	    data->flags |= MON_BUSY;
 	    data->checks = 0;
 	    gam_poll_add_missing(node);
-	    if (gam_node_is_dir(node))
-		trigger_dir_handler(gam_node_get_path(node), FALSE);
-	    else
-		trigger_file_handler(gam_node_get_path(node), FALSE);
+	    gam_poll_delist_node(node);
 	}
     }
 
     if ((event == 0) && (data->flags & MON_BUSY) && (data->checks > 10)) {
-#if 0
-	fprintf(stderr, "switching %s back to kernel monitoring\n", path);
-#endif
-	data->flags &= ~MON_BUSY;
-	data->checks = 0;
-	gam_poll_remove_missing(node);
-	if (gam_node_is_dir(node))
-	    trigger_dir_handler(gam_node_get_path(node), TRUE);
-	else
-	    trigger_file_handler(gam_node_get_path(node), TRUE);
+	if ((gam_node_get_subscriptions(node) != NULL) &&
+	    (!gam_exclude_check(data->path))) {
+	    GAM_DEBUG(DEBUG_INFO, "switching %s back to kernel monitoring\n",
+	              path);
+	    data->flags &= ~MON_BUSY;
+	    data->checks = 0;
+	    gam_poll_remove_missing(node);
+	    gam_poll_relist_node(node);
+	}
     }
 
     return(event);
 }
 
-static GList *
-list_intersection(GList * a, GList * b)
-{
-    GList *ret, *l;
-
-    ret = NULL;
-    for (l = a; l; l = l->next) {
-        if (g_list_find(b, l->data))
-            ret = g_list_prepend(ret, l->data);
-    }
-
-    return ret;
-}
-
 static void
-gam_poll_scan_directory_internal(GamNode * dir_node, GList * exist_subs,
-                                 gboolean scan_for_new)
+gam_poll_scan_directory_internal(GamNode * dir_node)
 {
     GDir *dir;
     const char *name, *dpath;
     char *path;
     GamNode *node;
     GaminEventType event = 0, fevent;
-    GList *dir_exist_subs = NULL;
     GList *children, *l;
     unsigned int exists = 0;
+    int is_dir_node;
 
-    g_return_if_fail(dir_node != NULL);
+    if (dir_node == NULL)
+        return;
 
     dpath = gam_node_get_path(dir_node);
 
-    g_return_if_fail(dpath != NULL);
-
-    if (!scan_for_new)
-        goto scan_files;
+    if (dpath == NULL)
+        return;
 
     if (!gam_node_get_subscriptions(dir_node))
         goto scan_files;
 
     event = poll_file(dir_node);
-    dir_exist_subs =
-        list_intersection(exist_subs,
-                          gam_node_get_subscriptions(dir_node));
 
-    if (event == 0 && !dir_exist_subs)
-        goto scan_files;
-
-    gam_poll_emit_event(dir_node, event, dir_exist_subs);
+    if (event != 0)
+	gam_poll_emit_event(dir_node, event, NULL);
 
     dir = g_dir_open(dpath, 0, NULL);
 
-    if (dir == NULL)
-        goto scan_files;
+    if (dir == NULL) {
+	GAM_DEBUG(DEBUG_INFO, "Poll: directory %s missing\n", dpath);
+	return;
+    }
 
     exists = 1;
 
-    gam_debug(DEBUG_INFO, "Poll: scanning directory %s\n", dpath);
+    GAM_DEBUG(DEBUG_INFO, "Poll: scanning directory %s\n", dpath);
     while ((name = g_dir_read_name(dir)) != NULL) {
         path = g_build_filename(gam_node_get_path(dir_node), name, NULL);
 
@@ -419,7 +491,6 @@ gam_poll_scan_directory_internal(GamNode * dir_node, GList * exist_subs,
             } else {
                 node = gam_node_new(path, NULL, TRUE);
                 gam_tree_add(tree, dir_node, node);
-
                 gam_node_set_flag(node, FLAG_NEW_NODE);
             }
         }
@@ -432,32 +503,27 @@ gam_poll_scan_directory_internal(GamNode * dir_node, GList * exist_subs,
 scan_files:
 
 
-    if (scan_for_new) {
-        if (exists)
-            gam_server_emit_event(dpath, GAMIN_EVENT_EXISTS, exist_subs);
-        else {
-            gam_server_emit_event(dpath, GAMIN_EVENT_DELETED, exist_subs);
-	}
-    }
+    is_dir_node = gam_node_is_dir(dir_node);
     children = gam_tree_get_children(tree, dir_node);
     for (l = children; l; l = l->next) {
+
         node = (GamNode *) l->data;
 
         fevent = poll_file(node);
 
-        if (gam_node_is_dir(node) &&
-            gam_node_has_flag(node, FLAG_NEW_NODE) &&
-            gam_node_get_subscriptions(node)) {
-            gam_node_unset_flag(node, FLAG_NEW_NODE);
-            gam_poll_scan_directory_internal(node, exist_subs,
-                                             scan_for_new);
-        } else if (gam_node_has_flag(node, FLAG_NEW_NODE)) {
-            gam_node_unset_flag(node, FLAG_NEW_NODE);
-            fevent = GAMIN_EVENT_CREATED;
-        }
+        if (gam_node_has_flag(node, FLAG_NEW_NODE)) {
+	    if (is_dir_node &&
+		gam_node_get_subscriptions(node)) {
+		gam_node_unset_flag(node, FLAG_NEW_NODE);
+		gam_poll_scan_directory_internal(node);
+	    } else {
+		gam_node_unset_flag(node, FLAG_NEW_NODE);
+		fevent = GAMIN_EVENT_CREATED;
+	    }
+	}
 
         if (fevent != 0) {
-            gam_poll_emit_event(node, fevent, exist_subs);
+            gam_poll_emit_event(node, fevent, NULL);
         } else {
 	    GamPollData *data;
 
@@ -466,17 +532,12 @@ scan_files:
 
 	    if ((data) && (!(data->flags & MON_MISSING)))
 		gam_server_emit_event(gam_node_get_path(node),
-				      GAMIN_EVENT_EXISTS, exist_subs);
-
+		                      gam_node_is_dir(node),
+				      GAMIN_EVENT_EXISTS, NULL, 0);
         }
     }
 
-    if (scan_for_new) {
-        gam_server_emit_event(dpath, GAMIN_EVENT_ENDEXISTS, exist_subs);
-    }
-
     g_list_free(children);
-    g_list_free(dir_exist_subs);
 }
 
 static gboolean
@@ -484,7 +545,9 @@ remove_directory_subscription(GamNode * node, GamSubscription * sub)
 {
     GList *children, *l;
     gboolean remove_dir;
-    
+
+    GAM_DEBUG(DEBUG_INFO, "remove_directory_subscription %s\n",
+              gam_node_get_path(node));
 
     node_remove_subscription(node, sub);
 
@@ -494,21 +557,15 @@ remove_directory_subscription(GamNode * node, GamSubscription * sub)
     for (l = children; l; l = l->next) {
         GamNode *child = (GamNode *) l->data;
 
-        if (gam_node_is_dir(child)) {
-            if (remove_directory_subscription(child, sub) && remove_dir) {
-                gam_tree_remove(tree, child);
-            } else {
-                remove_dir = FALSE;
-            }
-        } else {
-            /* node_remove_subscription(child, sub); */
-
-            if (!gam_node_get_subscriptions(child) && remove_dir) {
-                gam_tree_remove(tree, child);
-            } else {
-                remove_dir = FALSE;
-            }
-        }
+	if ((!gam_node_get_subscriptions(child)) && (remove_dir) &&
+	    (!gam_tree_has_children(tree, child))) {
+	    if (missing_resources != NULL) {
+		gam_poll_remove_missing (child);
+	    }
+	    gam_tree_remove(tree, child);
+	} else {
+	    remove_dir = FALSE;
+	}
     }
 
     g_list_free(children);
@@ -519,9 +576,12 @@ remove_directory_subscription(GamNode * node, GamSubscription * sub)
 static gboolean
 gam_poll_scan_callback(gpointer data) {
     int idx;
-
     static int in_poll_callback = 0;
 
+#ifdef VERBOSE_POLL
+    GAM_DEBUG(DEBUG_INFO, "gam_poll_scan_callback(): %d, %d items\n",
+              in_poll_callback, g_list_length(missing_resources));
+#endif
     if (in_poll_callback)
 	return(TRUE);
 
@@ -535,29 +595,43 @@ gam_poll_scan_callback(gpointer data) {
 	/*
 	 * do not simply walk the list as it may be modified in the callback
 	 */
-	G_LOCK(missing_resources);
 	node = (GamNode *) g_list_nth_data(missing_resources, idx);
-	G_UNLOCK(missing_resources);
 	
 	if (node == NULL) {
+#ifdef VERBOSE_POLL
+	    GAM_DEBUG(DEBUG_INFO, "  node %d == NULL\n", idx);
+#endif
 	    break;
 	} 
 	data = gam_node_get_data(node);
 	if (data == NULL) {
+#ifdef VERBOSE_POLL
+	    GAM_DEBUG(DEBUG_INFO, "  data %d == NULL\n", idx);
+#endif
 	    break;
 	} 
 
-	gam_poll_scan_directory_internal(node, NULL, TRUE);
+#ifdef VERBOSE_POLL
+	GAM_DEBUG(DEBUG_INFO, "Checking missing file %s", data->path);
+#endif
+	if (node->is_dir) {
+	    gam_poll_scan_directory_internal(node);
+	} else {
+	    GaminEventType event;
+
+	    event = poll_file(node);
+	    gam_poll_emit_event(node, event, NULL);
+	}
+
 	/*
 	 * if the resource exists again and is not in a special monitoring
 	 * mode then switch back to dnotify for monitoring.
 	 */
-	if (data->flags == 0) {
+	if ((data->flags == 0) && (!gam_exclude_check(data->path))) {
 	    gam_poll_remove_missing(node);
-	    if (gam_node_is_dir(node))
-		trigger_dir_handler(gam_node_get_path(node), TRUE);
-	    else
-		trigger_file_handler(gam_node_get_path(node), TRUE);
+	    if (gam_node_get_subscriptions(node) != NULL) {
+	        gam_poll_relist_node(node);
+	    }
 	}
     }
 
@@ -576,20 +650,104 @@ prune_tree(GamNode * node)
         !gam_node_get_subscriptions(node)) {
         GamNode *parent;
 
+	GAM_DEBUG(DEBUG_INFO,
+		  "prune_tree: node %s\n",
+		  gam_node_get_path(node));
+
         parent = gam_node_parent(node);
-        gam_tree_remove(tree, node);
+        if (missing_resources != NULL) {
+		gam_poll_remove_missing(node);
+        }
+        if (all_resources != NULL) {
+                all_resources = g_list_remove (all_resources, node);
+        }
+	gam_tree_remove(tree, node);
         prune_tree(parent);
     }
 }
 
 
+static gboolean
+gam_poll_scan_all_callback(gpointer data) {
+    int idx;
+    GList *subs, *l;
+    GamNode *node;
+
+    static int in_poll_callback = 0;
+
+    if (in_poll_callback)
+	return(TRUE);
+ 
+    in_poll_callback++;
+
+    if (new_subs != NULL) {
+        /* we don't want to block the main loop */
+        subs = new_subs;
+        new_subs = NULL;
+
+        GAM_DEBUG(DEBUG_INFO,
+                  "%d new subscriptions.\n", g_list_length(subs));
+
+        for (l = subs; l; l = l->next) {
+            GamSubscription *sub = l->data;
+
+            const char *path = gam_subscription_get_path(sub);
+
+            node = gam_tree_get_at_path(tree, path);
+            if (!node) {
+                node = gam_tree_add_at_path(tree, path,
+                                            gam_subscription_is_dir(sub));
+            }
+
+            if (node_add_subscription(node, sub) < 0) {
+                gam_error(DEBUG_INFO,
+                          "Failed to add subscription for: %s\n", path);
+	    }
+    	    if (!gam_node_is_dir(node)) {
+		    char *parent;
+		    
+		    parent = g_path_get_dirname (path);
+		    node = gam_tree_get_at_path(tree, parent);
+	            if (!node) {
+	                node = gam_tree_add_at_path(tree, parent,
+                                            gam_subscription_is_dir(sub));
+        	    }
+		    g_free (parent);
+	    }
+
+	      if (g_list_find(all_resources, node) == NULL) {
+		all_resources = g_list_prepend (all_resources, node);
+	      }
+    	}
+	g_list_free (subs);
+    }
+
+    current_time = time(NULL);
+    for (idx = 0;;idx++) {
+	
+	/*
+	 * do not simply walk the list as it may be modified in the callback
+	 */
+	node = (GamNode *) g_list_nth_data(all_resources, idx);
+
+	if (node == NULL) {
+	    break;
+	} 
+
+	gam_poll_scan_directory_internal(node);
+    }
+
+    in_poll_callback = 0;
+    return(TRUE);
+}
+
 
 /**
  * @defgroup Polling Polling Backend
  * @ingroup Backends
  * @brief Polling backend API
  *
- * This is the default backend used in Marmot.  It basically just calls
+ * This is the default backend used in Gamin.  It basically just calls
  * stat() on files/directories every so often to see when things change.  The
  * statting happens in a separate thread, controllable with arguments to
  * #gam_poll_init_full().
@@ -609,11 +767,13 @@ gam_poll_add_missing(GamNode *node) {
 #if 0
     fprintf(stderr, "Adding %s to polling\n", gam_node_get_path(node));
 #endif
-    gam_debug(DEBUG_INFO, "Poll adding missing node %s\n",
+    GAM_DEBUG(DEBUG_INFO, "Poll adding missing node %s\n",
               gam_node_get_path(node));
-    G_LOCK(missing_resources);
-    missing_resources = g_list_prepend(missing_resources, node);
-    G_UNLOCK(missing_resources);
+    if (g_list_find(missing_resources, node) == NULL) {
+	missing_resources = g_list_prepend(missing_resources, node);
+    } else {
+	GAM_DEBUG(DEBUG_INFO, "  already registered\n");
+    }
 }
 
 /**
@@ -627,11 +787,9 @@ gam_poll_remove_missing(GamNode *node) {
 #if 0
     fprintf(stderr, "Removing %s from polling\n", gam_node_get_path(node));
 #endif
-    gam_debug(DEBUG_INFO, "Poll removing missing node %s\n",
+    GAM_DEBUG(DEBUG_INFO, "Poll removing missing node %s\n",
               gam_node_get_path(node));
-    G_LOCK(missing_resources);
     missing_resources = g_list_remove_all(missing_resources, node);
-    G_UNLOCK(missing_resources);
 }
 
 /**
@@ -652,6 +810,7 @@ gam_poll_init_full(gboolean start_scan_thread)
         g_timeout_add(1000, gam_poll_scan_callback, NULL);
 	poll_mode = 1;
     } else {
+        g_timeout_add(1000, gam_poll_scan_all_callback, NULL);
 	poll_mode = 2;
     }
     tree = gam_tree_new();
@@ -660,7 +819,7 @@ gam_poll_init_full(gboolean start_scan_thread)
     gam_backend_remove_subscription = gam_poll_remove_subscription;
     gam_backend_remove_all_for = gam_poll_remove_all_for;
 
-    gam_debug(DEBUG_INFO, "Initialized Poll\n");
+    GAM_DEBUG(DEBUG_INFO, "Initialized Poll\n");
     return TRUE;
 }
 
@@ -690,27 +849,85 @@ gam_poll_add_subscription(GamSubscription * sub)
     const char *path;
     gboolean is_dir;
 
+    if (g_list_find(new_subs, sub))
+        return FALSE;
+
     path = gam_subscription_get_path(sub);
     is_dir = gam_subscription_is_dir(sub);
 
-/***
-    node = gam_tree_get_at_path(tree, path);
-
-    if (!node) {
-        node = gam_tree_add_at_path(tree, path, is_dir);
-    }
- ***/
-
     gam_listener_add_subscription(gam_subscription_get_listener(sub), sub);
 
-    G_LOCK(new_subs);
     new_subs = g_list_prepend(new_subs, sub);
-    G_UNLOCK(new_subs);
 
-    gam_debug(DEBUG_INFO, "Poll: added subscription\n");
+    GAM_DEBUG(DEBUG_INFO, "Poll: added subscription\n");
     return TRUE;
 }
 
+/**
+ * gam_poll_remove_subscription_real:
+ * @sub: a subscription
+ *
+ * Implements the removal of a subscription, including
+ * trimming the tree and deactivating the kernel back-end if needed.
+ */
+static void
+gam_poll_remove_subscription_real(GamSubscription * sub)
+{
+    GamNode *node;
+
+    node = gam_tree_get_at_path(tree, gam_subscription_get_path(sub));
+
+    if (node != NULL) {
+        if (!gam_node_is_dir(node)) {
+            GAM_DEBUG(DEBUG_INFO, "Removing node sub: %s\n",
+                      gam_subscription_get_path(sub));
+            node_remove_subscription(node, sub);
+
+            if (!gam_node_get_subscriptions(node)) {
+                GamNode *parent;
+
+                if (missing_resources != NULL) {
+                    gam_poll_remove_missing(node);
+                }
+                if (all_resources != NULL) {
+                    all_resources = g_list_remove(all_resources, node);
+                }
+                if (gam_tree_has_children(tree, node)) {
+                    fprintf(stderr,
+                            "node %s is not dir but has children\n",
+                            gam_node_get_path(node));
+                } else {
+                    parent = gam_node_parent(node);
+                    gam_tree_remove(tree, node);
+
+                    prune_tree(parent);
+                }
+            }
+        } else {
+            GAM_DEBUG(DEBUG_INFO, "Removing directory sub: %s\n",
+                      gam_subscription_get_path(sub));
+            if (remove_directory_subscription(node, sub)) {
+                GamNode *parent;
+
+                if (missing_resources != NULL) {
+                    gam_poll_remove_missing(node);
+                }
+                if (all_resources != NULL) {
+                    all_resources = g_list_remove(all_resources, node);
+                }
+                parent = gam_node_parent(node);
+		if (!gam_tree_has_children(tree, node)) {
+		    gam_tree_remove(tree, node);
+		}
+
+                prune_tree(parent);
+            }
+        }
+    }
+
+    gam_subscription_free(sub);
+}
+
 /**
  * Removes a subscription which was being polled.
  *
@@ -725,12 +942,10 @@ gam_poll_remove_subscription(GamSubscription * sub)
     /*
      * make sure the subscription still isn't in the new subscription queue
      */
-    G_LOCK(new_subs);
     if (g_list_find(new_subs, sub)) {
-        gam_debug(DEBUG_INFO, "new subscriptions is removed\n");
+        GAM_DEBUG(DEBUG_INFO, "new subscriptions is removed\n");
         new_subs = g_list_remove_all(new_subs, sub);
     }
-    G_UNLOCK(new_subs);
 
     node = gam_tree_get_at_path(tree, gam_subscription_get_path(sub));
     if (node == NULL) {
@@ -743,11 +958,13 @@ gam_poll_remove_subscription(GamSubscription * sub)
     gam_listener_remove_subscription(gam_subscription_get_listener(sub),
                                      sub);
 
-    G_LOCK(removed_subs);
-    removed_subs = g_list_prepend(removed_subs, sub);
-    G_UNLOCK(removed_subs);
+    GAM_DEBUG(DEBUG_INFO, "Tree has %d nodes\n",
+	      gam_tree_get_size(tree));
+    gam_poll_remove_subscription_real(sub);
+    GAM_DEBUG(DEBUG_INFO, "Tree has %d nodes\n",
+	      gam_tree_get_size(tree));
 
-    gam_debug(DEBUG_INFO, "Poll: removed subscription\n");
+    GAM_DEBUG(DEBUG_INFO, "Poll: removed subscription\n");
     return TRUE;
 }
 
@@ -783,16 +1000,13 @@ gam_poll_remove_all_for(GamListener * listener)
  * Scans a directory for changes, and emits events if needed.
  *
  * @param path the path to the directory to be scanned
- * @param exist_subs a list of type #GamSubscription of new subscriptions
- * which need to be sent the EXIST event.
  */
 void
-gam_poll_scan_directory(const char *path, GList * exist_subs)
+gam_poll_scan_directory(const char *path)
 {
     GamNode *node;
 
-    gam_debug(DEBUG_INFO, "Poll: scanning %s: subs %d\n",
-              path, exist_subs != NULL);
+    GAM_DEBUG(DEBUG_INFO, "Poll: directory scanning %s\n", path);
 
     current_time = time(NULL);
     node = gam_tree_get_at_path(tree, path);
@@ -803,8 +1017,103 @@ gam_poll_scan_directory(const char *path, GList * exist_subs)
 	return;
     }
 
-    gam_poll_scan_directory_internal(node, exist_subs, TRUE);
-    gam_debug(DEBUG_INFO, "Poll: scanning %s done\n", path);
+    gam_poll_scan_directory_internal(node);
+    GAM_DEBUG(DEBUG_INFO, "Poll: scanning %s done\n", path);
+}
+
+/**
+ * First dir scanning on a new subscription, generates the Exists EndExists
+ * events.
+ */
+static void
+gam_poll_first_scan_dir(GamSubscription *sub, GamNode *dir_node,
+                        GamPollData *data, const char *dpath)
+{
+    GDir *dir;
+    char *path;
+    GList * subs;
+    int with_exists = 1;
+    const char *name;
+    GamNode *node;
+
+    GAM_DEBUG(DEBUG_INFO, "Looking for existing files in: %s...\n", dpath);
+
+    if (gam_subscription_has_option(sub, GAM_OPT_NOEXISTS)) {
+        with_exists = 0;
+	GAM_DEBUG(DEBUG_INFO, "   Exists not wanted\n");
+    }
+
+    subs = g_list_prepend(NULL, sub);
+
+    dir = g_dir_open(dpath, 0, NULL);
+
+    if (dir == NULL) {
+	GAM_DEBUG(DEBUG_INFO, "Monitoring missing dir: %s\n", dpath);
+	gam_server_emit_event(dpath, 1, GAMIN_EVENT_DELETED, subs, 1);
+	node = gam_node_new(dpath, NULL, TRUE);
+	if (node == NULL) {
+	    gam_error(DEBUG_INFO, "Failed to allocate node for: %s\n", dpath);
+	    goto done;
+	}
+	data = gam_poll_data_new(dpath);
+	if (data == NULL) {
+	    gam_error(DEBUG_INFO, "Failed to allocate data for: %s\n", dpath);
+	    goto done;
+	}
+	gam_node_add_subscription(node, sub);
+	gam_node_set_data(node, data,
+			  (GDestroyNotify) gam_poll_data_destroy);
+	if (g_file_test(dpath, G_FILE_TEST_EXISTS)) {
+	    data->flags = MON_WRONG_TYPE;
+	    node->is_dir = 0;
+	} else {
+	    data->flags = MON_MISSING;
+	    gam_poll_add_missing(node);
+	}
+	goto done;
+    } 
+
+    if (with_exists)
+	gam_server_emit_event(dpath, 1, GAMIN_EVENT_EXISTS, subs, 1);
+    while ((name = g_dir_read_name(dir)) != NULL) {
+        path = g_build_filename(dpath, name, NULL);
+
+        node = gam_tree_get_at_path(tree, path);
+
+        if (!node) {
+	    GAM_DEBUG(DEBUG_INFO, "Unregistered node %s\n", path);
+            if (!g_file_test(path, G_FILE_TEST_IS_DIR)) {
+                node = gam_node_new(path, NULL, FALSE);
+            } else {
+                node = gam_node_new(path, NULL, TRUE);
+            }
+	    data = gam_poll_data_new(path);
+	    if (data != NULL) {
+		gam_node_set_data(node, data,
+				  (GDestroyNotify) gam_poll_data_destroy);
+
+		stat(data->path, &(data->sbuf));
+		gam_node_set_is_dir(node, (S_ISDIR(data->sbuf.st_mode) != 0));
+		if (gam_exclude_check(path))
+		    data->flags |= MON_NOKERNEL;
+	    }
+	    gam_tree_add(tree, dir_node, node);
+        }
+	if (with_exists)
+	    gam_server_emit_event(name, 1, GAMIN_EVENT_EXISTS, subs, 1);
+
+        g_free(path);
+    }
+
+    g_dir_close(dir);
+
+done:
+    if (with_exists)
+	gam_server_emit_event(dpath, 1, GAMIN_EVENT_ENDEXISTS, subs, 1);
+
+    g_list_free(subs);
+
+    GAM_DEBUG(DEBUG_INFO, "Done scanning %s\n", dpath);
 }
 
 /**
@@ -821,22 +1130,21 @@ gam_poll_consume_subscriptions(void)
      * (specifically, sending them the EXIST event)
      */
     current_time = time(NULL);
-    G_LOCK(new_subs);
     if (new_subs != NULL) {
         /* we don't want to block the main loop */
         subs = new_subs;
         new_subs = NULL;
-        G_UNLOCK(new_subs);
 
-        gam_debug(DEBUG_INFO,
+        GAM_DEBUG(DEBUG_INFO,
                   "%d new subscriptions.\n", g_list_length(subs));
 
         for (l = subs; l; l = l->next) {
             GamSubscription *sub = l->data;
             GamNode *node;
-	    GamPollData *data;
+            GamPollData *data;
+            int node_is_dir;
 
-	    const char *path = gam_subscription_get_path(sub);
+            const char *path = gam_subscription_get_path(sub);
 
             node = gam_tree_get_at_path(tree, path);
             if (!node) {
@@ -844,100 +1152,42 @@ gam_poll_consume_subscriptions(void)
                                             gam_subscription_is_dir(sub));
             }
 
-            node_add_subscription(node, sub);
+            if (node_add_subscription(node, sub) < 0) {
+                gam_error(DEBUG_INFO,
+                          "Failed to add subscription for: %s\n", path);
+	        return;
+	    }
 
-            if (gam_node_is_dir(node)) {
-
-                gam_debug(DEBUG_INFO,
-                          "Looking for existing files in: %s...\n",
-                          path);
-
-                gam_poll_scan_directory_internal(node, subs, TRUE);
-
-                gam_debug(DEBUG_INFO, "Done scanning %s\n", path);
+            node_is_dir = gam_node_is_dir(node);
+            if (node_is_dir) {
+                gam_poll_first_scan_dir(sub, node, data, path);
             } else {
-		GaminEventType event;
+                GaminEventType event;
 
                 event = poll_file(node);
-                gam_debug(DEBUG_INFO,
-                          "New file subscription: %s event %d\n", path, event);
-		if ((event == 0) || (event == GAMIN_EVENT_EXISTS) ||
-		    (event == GAMIN_EVENT_CHANGED) ||
-		    (event == GAMIN_EVENT_CREATED)) {
-		    gam_server_emit_one_event(path, GAMIN_EVENT_EXISTS, sub);
-		} else if (event != 0) {
-		    gam_server_emit_one_event(path, GAMIN_EVENT_DELETED, sub);
-		}
-		gam_server_emit_one_event(path, GAMIN_EVENT_ENDEXISTS, sub);
-	    }
-	    data = gam_node_get_data(node);
-	    if ((data) && ((data->flags & MON_MISSING) ||
-	                   (data->flags & MON_NOKERNEL))) {
-		gam_poll_add_missing(node);
-	    }
+                GAM_DEBUG(DEBUG_INFO,
+                          "New file subscription: %s event %d\n", path,
+                          event);
+                if ((event == 0) || (event == GAMIN_EVENT_EXISTS)
+                    || (event == GAMIN_EVENT_CHANGED)
+                    || (event == GAMIN_EVENT_CREATED)) {
+                    gam_server_emit_one_event(path, node_is_dir,
+                                              GAMIN_EVENT_EXISTS, sub, 0);
+                } else if (event != 0) {
+                    gam_server_emit_one_event(path, node_is_dir,
+                                              GAMIN_EVENT_DELETED, sub, 0);
+                }
+                gam_server_emit_one_event(path, node_is_dir,
+                                          GAMIN_EVENT_ENDEXISTS, sub, 0);
+            }
+            data = gam_node_get_data(node);
+            if ((data) && ((data->flags & MON_MISSING) ||
+                           (data->flags & MON_NOKERNEL))) {
+                gam_poll_add_missing(node);
+            }
         }
 
         g_list_free(subs);
-    } else {
-        G_UNLOCK(new_subs);
-    }
-
-    /* check for things that have been removed, and remove them */
-    G_LOCK(removed_subs);
-    if (removed_subs != NULL) {
-        subs = removed_subs;
-        removed_subs = NULL;
-        G_UNLOCK(removed_subs);
-
-        gam_debug(DEBUG_INFO, "Tree has %d nodes\n",
-                  gam_tree_get_size(tree));
-        for (l = subs; l; l = l->next) {
-            GamSubscription *sub = l->data;
-            GamNode *node = gam_tree_get_at_path(tree,
-                                                 gam_subscription_get_path
-                                                 (sub));
-
-            gam_debug(DEBUG_INFO, "Removing: %s\n",
-                      gam_subscription_get_path(sub));
-            if (node != NULL) {
-		if (!gam_node_is_dir(node)) {
-		    node_remove_subscription(node, sub);
-
-		    if (!gam_node_get_subscriptions(node)) {
-			GamNode *parent;
-
-			if (missing_resources != NULL) {
-			    gam_poll_remove_missing(node);
-			}
-			parent = gam_node_parent(node);
-			gam_tree_remove(tree, node);
-
-			prune_tree(parent);
-		    }
-		} else {
-		    if (remove_directory_subscription(node, sub)) {
-			GamNode *parent;
-
-			if (missing_resources != NULL) {
-			    gam_poll_remove_missing(node);
-			}
-			parent = gam_node_parent(node);
-			gam_tree_remove(tree, node);
-
-			prune_tree(parent);
-		    }
-		}
-	    }
-
-            gam_subscription_free(sub);
-        }
-        g_list_free(subs);
-
-        gam_debug(DEBUG_INFO, "Tree has %d nodes\n",
-                  gam_tree_get_size(tree));
-
-    } else {
-        G_UNLOCK(removed_subs);
     }
 }
 
